# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import json
import pytest # noqa needed for capsys

from crush import Crush


class TestCrush(object):

    def test_parse_empty(self, capsys):
        empty = {
            'trees': {"dc1": {
                '~type~': 'root',
            }}
        }
        assert Crush(verbose=True).parse(empty)
        out, err = capsys.readouterr()
        assert 'trees' in out
        assert '~type~' in out

        assert Crush().parse(empty)
        out, err = capsys.readouterr()
        assert 'trees' not in out

    def test_map(self):
        map = """
        {
          "trees": { "dc1": {
           "~type~": "root",
           "host0": {
            "~type~": "host",
            "device0": { "~id~": 0, "~weight~": 1.0 },
            "device1": { "~id~": 1, "~weight~": 2.0 }
           },
           "host1": {
            "~type~": "host",
            "device2": { "~id~": 2, "~weight~": 1.0 },
            "device3": { "~id~": 3, "~weight~": 2.0 }
           },
           "host2": {
            "~type~": "host",
            "device4": { "~id~": 4, "~weight~": 1.0 },
            "device5": { "~id~": 5, "~weight~": 2.0 }
           }
          } },
          "rules": {
           "data": [
              [ "take", "dc1" ],
              [ "chooseleaf", "firstn", 0, "type", "host" ],
              [ "emit" ]
           ]
          }
        }

        """
        c = Crush(verbose=1)
        assert c.parse(json.loads(map))
        assert len(c.map(rule="data", value=1234, replication_count=1)) == 1

# Local Variables:
# compile-command: "cd .. ; tox -e py27 tests/test_crush.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
try:  # chicken and egg problem when running pip install -e . from sources
    from crush.libcrush import LibCrush
except:
    pass


class Crush(object):
    """Control object placement in a hierarchy.

    The algorithms optimize the placement so that:

    - the devices are filled according to their weight

    - the number of objects that move is proportional to the magnitude
      of the weight of the devices removed or added

    """

    def __init__(self, verbose=False):
        """Create a Crush.

        If the optional argument `verbose` is set to True, all methods
        will print debug information on stdout. It defaults to False.

        """
        self.c = LibCrush(verbose and 1 or 0)

    def parse(self, crushmap):
        """Validate and parse the `crushmap` object.

        The `crushmap` is a hierarchical description of devices in
        which objects can be stored and rules to place the objects.
        It is verified to obey the specifications below. An exception
        is raised on the first error found.
        ::

            crushmap = {
              # optional (default: none)
              "trees": trees,

              # optional (default: none)
              "rules": rules,
            }

        The "trees" are the roots of device hierarchies, the "rules" describe
        various object placement strategies for this device hierarchy.
        ::

            trees = {
              # optional (default: none)
              <root name str>: bucket or device,
              <root name str>: bucket or device,
              ...
            }

        The **root name** is the name of a top level bucket or device.
        ::

            bucket = {
              # mandatory
              "~type~": <str>,

              # optional (default: first available id)
              "~id~": <negative int>,

              # optional (default: "straw2")
              "~algorithm~": "uniform" or "list" or "straw2",

              # optional (default: cumulated children weights or none)
              "~weight~": <postive float>,

              # optional (default: none)
              <child name str>: bucket or device,
              <child name str>: bucket or device,
              ...
            }

        The **~type~** is a user defined string that can be used by
        **rules** to select all buckets of the same type.

        The **~id~** must either be set for all buckets or not at
        all. If the **~id~** is provided, it must be a unique negative
        number. If it is not provided, the first available id is
        used.

        The **~weight~** must either be set for all buckets or not at
        all. If not set, **~weight~** defaults to the cumulated weight
        of the immediate children bucket or devices, recursively,
        bottom to top.

        Children within a bucket are chosen with one of three
        **~algorithms~** representing a tradeoff between performance
        and reorganization efficiency. If you are unsure, we recommend
        using **"straw2"**. The table summarizes how the speed of each
        option measures up against mapping stability when items are
        added or removed::

            Bucket Alg     Speed       Additions    Removals
            ------------------------------------------------
            "straw2"        O(n)       optimal      optimal
            "uniform"       O(1)       poor         poor
            "list"          O(n)       optimal      poor

        - **"straw2"**: List and tree buckets are structured such that
          a limited number of hash values need to be calculated and
          compared to weights in order to select a bucket child. In
          doing so, they divide and conquer in a way that either gives
          certain children precedence (e. g., those at the beginning
          of a list) or obviates the need to consider entire subtrees
          of children at all. That improves the performance of the
          replica placement process, but can also introduce suboptimal
          reorganization behavior when the contents of a bucket change
          due an addition, removal, or re-weighting of an child. The
          straw2 bucket type allows all children to fairly `compete`
          against each other for replica placement through a process
          analogous to a draw of straws. To place a replica, a straw
          of random length is drawn for each child in the bucket.  The
          child with the longest straw wins.  The length of each straw
          is initially a value in a fixed range.  Each straw length is
          scaled by a factor based on the child’s weight so that
          heavily weighted children are more likely to win the draw.
          Although this process is almost twice as slow (on average)
          than a list bucket and even slower than a tree bucket (which
          scales logarithmically), straw2 buckets result in optimal
          data movement between nested children when modified.

        - **"uniform"**: Devices are rarely added individually in a
          large system.  Instead, new storage is typically deployed in
          blocks of identical devices, often as an additional shelf in
          a server rack or perhaps an entire cabinet. Devices reaching
          their end of life are often similarly decommissioned as a
          set (individual failures aside), making it natural to treat
          them as a unit. Uniform buckets are used to represent an
          identical set of devices in such circumstances. The key
          advantage in doing so is performance related: Crush can map
          replicas into uniform buckets in constant time. In cases
          where the uniformity restrictions are not appropriate, other
          bucket types can be used. If the size of a uniform bucket
          changes, there is a complete reshuffling of data between
          devices, much like conventional hash-based distribution
          strategies.

        - **"list"**: List buckets structure their contents as a
          linked list, and can contain children with arbitrary
          weights.  To place a replica, Crush begins at the head of
          the list with the most recently added child and compares its
          weight to the sum of all remaining children’ weights.
          Depending on the value of the hash function, either the
          current child is chosen with the appropriate probability, or
          the process continues recursively down the list. This is a
          natural and intuitive choice for an expanding cluster:
          either an object is relocated to the newest device with some
          appropriate probability, or it remains on the older devices
          as before.  The result is optimal data migration when
          children are added to the bucket. Children removed from the
          middle or tail of the list, however, can result in a
          significant amount of unnecessary movement, making list
          buckets most suitable for circumstances in which they never
          (or very rarely) shrink.

        There can be many children or none.
        ::

            device = {
              # mandatory
              "~id~": <positive int>,

              # optional (default: 1.0)
              "~weight~": <postive float>,
            }

        The **~id~** must be a unique positive number.

        If the **~weight~** of a device A is lower than the
        **~weight~** of a device B, it will be less likely to be used.
        A common pattern is to set the **~weight~** to 2.0 for 2TB
        devices, 1.0 for 1TB devices, 0.5 for 500GB devices, etc.
        ::

            rules = {
              # optional (default: none)
              <rule name str>: rule,
              <rule name str>: rule,
              ...
            }

        A **rule** maps an object (see the map function) to a list
        of devices. There can be multiple rules depending on the
        mapping strategy.
        ::

            rule = [ step, step, ... ]

        The **rule** interprets each **step** in sequence and the last
        one must be **"emit"**.
        ::

            step = [ "take", <bucket name str> ]

        Select the **bucket name**.
        ::

            step = [
              "choose_firstn" or "choose_indep",
              <replication count positive int>,
              <bucket type str>
            ]

        Recursively explore each bucket currently selected, looking for
        **replication count** buckets of the required **bucket type**
        and select them.

        If **replication count** is zero, the number of buckets
        to select will be determined by the `replication_count` argument of
        the `map` method, i.e. **replication count** is set to
        match the desired number of replicas.
        ::

            step = [
              "chooseleaf_firstn" or "chooseleaf_indep",
              <replication count positive int>,
              <bucket type str>
            ]

        Recursively explore each bucket currently selected, looking for
        **replication count** devices within all buckets of
        the required **bucket type** and select them.

        If **replication count** is zero, the number of devices
        to select will be determined by the `replication_count` argument of
        the `map` method, i.e. **replication count** is set to
        match the desired number of replicas.
        ::

            step = [ "emit" ]

        Append the selection to the results and clear the selection.
        """
        self.c.parse(crushmap)
        return True

    def map(self, rule, value, replication_count, weights=None):
        """Map an object to a list of devices.

        The **rule** is used to map the **value** (representing an
        object) to the desired number of devices
        (**replication_count**) and return them in a list. The
        probabilities for a given device to be selected can be
        modified by the **weights** dictionnary.

        If the mapping is successful, a list of device containing
        exactly **replication_count** devices is returned. If the
        mapping fails, the list may contains less devices or some
        names may be replaced by None. For instance, if asking
        for 3 replicas, the result of a failed mapping may be::

            [ "device1", "device5" ] # 2 instead of 3
            [ "device8", None, "device0" ] # second device is missing

        The dictionary of **weights** modifies the probabilities for
        the device names it contains. The value is a float in the
        range [0..1]. If the weight is 0, the device will never be
        selected.  If the weight is 1, the probability that the device
        is selected is not modified. If a device is not in the
        **weights** dictionary, its probability is not modified. For
        example::

            { "device0": 0.50, "device1": 0.75 }

        will multiply the probability of "device0" by 0.50 (i.e,
        reduce it by 50%) and "device1" by 0.75 (i.e, reduce it by
        25%).

        - **rule**: the rule name (required string)

        - **value**: the number to map (required integer)

        - **replication_count**: the desired number of devices
            (required positive integer)

        - **weights**: map of name to weight float (optional, default to None)

        Return a list of device names.

        """
        kwargs = {
            "rule": rule,
            "value": value,
            "replication_count": replication_count,
        }
        if weights:
            kwargs["weight"] = weights
        return self.c.map(**kwargs)



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import json
import pytest

from crush.libcrush import LibCrush


class TestLibCrush(object):

    def test_parse_verbose(self, capsys):
        empty = {
            'trees': {"dc1": {
                '~type~': 'root',
            }}
        }
        assert LibCrush(verbose=1).parse(empty)
        out, err = capsys.readouterr()
        assert 'trees' in out
        assert '~type~' in out

        assert LibCrush().parse(empty)
        out, err = capsys.readouterr()
        assert 'trees' not in out

    def test_parse_empty(self):
        LibCrush().parse({})

    def test_parse_no_argument(self):
        with pytest.raises(TypeError):
            LibCrush().parse()

    def test_parse_argument_wrong(self):
        with pytest.raises(TypeError):
            LibCrush().parse([])

    def test_parse_trees_type_wrong(self):
        wrong = {
            'trees': []
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a dict' in str(e.value)

    def test_parse_invalid_algorithm(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                '~algorithm~': 'FOOBAR',
            }}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'not FOOBAR' in str(e.value)
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                '~algorithm~': 0
            }}
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

    def test_parse_trees_invalid_key(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                1: 'some',
            }}
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_bucket_invalid_key(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                '~INVALID': 1,
            }}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert '~INVALID is not' in str(e.value)

    def test_parse_device_invalid_key(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                'device0': {
                    '~id~': 1,
                    'INVALID': 1,
                }
            }}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "'INVALID' is not" in str(e.value)

    def test_parse_invalid_type(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 1,
            }}
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_invalid_id(self):
        wrong = {
            'trees': {"dc1": {
                '~id~': "some",
            }}
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_device_id_invalid(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                'device0': {
                    '~id~': -1,
                }
            }}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a positive integer" in str(e.value)

    def test_parse_bucket_id_invalid(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                '~id~': 2,
            }}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be a negative integer" in str(e.value)

    def test_parse_bucket_duplicate_id(self):
        duplicate_id = {
            'trees': {"dc1": {
                '~type~': 'root',
                '~id~': -1,
                'host0': {
                    '~type~': 'host',
                    '~id~': -2,
                },
                'host1': {
                    '~type~': 'host',
                    '~id~': -2,
                },
            }}
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(duplicate_id)
        assert ' -17 ' in str(e.value)

    def test_parse_weight_invalid(self):
        wrong = {
            'trees': {"dc1": {
                '~type~': 'root',
                '~weight~': "some",
            }}
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_rules_type_wrong(self):
        wrong = {
            'rules': []
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be a dict' in str(e.value)

    def test_parse_rules_invalid_key(self):
        wrong = {
            'rules': {
                1: 'root',
            }
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_step_missing_operand(self):
        wrong = {
            'rules': {
                'data': [
                    []
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'missing operand' in str(e.value)

    def test_parse_step_unknown_operand(self):
        wrong = {
            'rules': {
                'data': [
                    ['unknown']
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'operand unknown' in str(e.value)

    def test_parse_step_operand_bad_type(self):
        wrong = {
            'rules': {
                'data': [
                    [1]
                ]
            }
        }
        with pytest.raises(TypeError):
            LibCrush().parse(wrong)

    def test_parse_step_take_bad(self):
        wrong = {
            'rules': {
                'data': [
                    ["take"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly two' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["take", 1]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["take", u"unknown"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'not a known bucket' in str(e.value)

    def test_parse_step_emit_bad(self):
        wrong = {
            'rules': {
                'data': [
                    ["emit", 2]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly one' in str(e.value)

    def test_parse_step_set_bad(self):
        wrong = {
            'rules': {
                'data': [
                    ["set_choose_tries"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly two' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["set_wrong", 1]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'set operand unknown' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["set_choose_tries", "LK"]
                ]
            }
        }
        with pytest.raises(Exception) as e:
            LibCrush().parse(wrong)

    def test_parse_step_choose(self):
        wrong = {
            'rules': {
                'data': [
                    ["choose"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'exactly five' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", 1, 2, 3, 4]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "notgood", 2, 3, 4]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'unknown notgood' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", "notgood", 3, 4]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", -1, 3, 4]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert 'must be positive' in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", 0, "notgood", 4]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "must be 'type'" in str(e.value)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", 0, "type", 4]
                ]
            }
        }
        with pytest.raises(TypeError) as e:
            LibCrush().parse(wrong)

        wrong = {
            'rules': {
                'data': [
                    ["choose", "firstn", 0, "type", "unknown"]
                ]
            }
        }
        with pytest.raises(RuntimeError) as e:
            LibCrush().parse(wrong)
        assert "type unknown is unknown" in str(e.value)

    def test_parse_various_ok(self):
        map = """
        {
          "trees": { "dc1": {
           "~id~": -1,
           "~type~": "root",
           "~algorithm~": "list",
           "rack0": {
            "~id~": -2,
            "~type~": "rack",
            "host0": {
             "~id~": -3,
             "~type~": "host",
             "device0": { "~id~": 0, "~weight~": 1.0 },
             "device2": { "~id~": 1, "~weight~": 2.0 }
            },
            "host1": {
             "~id~": -4,
             "~type~": "host",
             "device3": { "~id~": 2, "~weight~": 2.0 },
             "device4": { "~id~": 3, "~weight~": 2.0 }
            }
           },
           "rack1": {
            "~id~": -5,
            "~type~": "rack",
            "host2": {
             "~id~": -6,
             "~type~": "host",
             "device5": { "~id~": 4, "~weight~": 1.0 },
             "device6": { "~id~": 5, "~weight~": 1.0 }
            },
            "host3": {
             "~id~": -7,
             "~type~": "host",
             "device7": { "~id~": 6, "~weight~": 1.0 },
             "device8": { "~id~": 7, "~weight~": 1.0 }
            }
           }
          } },
          "rules": {
           "data": [
             [ "take", "dc1" ],
             [ "chooseleaf", "firstn", 0, "type", "host" ],
             [ "emit" ]
           ]
          }
        }

        """
        assert LibCrush(verbose=1).parse(json.loads(map))

    def test_map_ok(self):
        map = """
        {
          "trees": {
            "dc1": {
              "~type~": "root",
              "~id~": -1,
              "host0": {
                "~type~": "host",
                "~id~": -2,
                "device0": { "~id~": 0, "~weight~": 1.0 },
                "device1": { "~id~": 1, "~weight~": 2.0 }
              },
              "host1": {
                "~type~": "host",
                "~id~": -3,
                "device2": { "~id~": 2, "~weight~": 1.0 },
                "device3": { "~id~": 3, "~weight~": 2.0 }
              },
              "host2": {
                "~type~": "host",
                "~id~": -4,
                "device4": { "~id~": 4, "~weight~": 1.0 },
                "device5": { "~id~": 5, "~weight~": 2.0 }
              }
            }
          },
          "rules": {
            "data": [
              [ "take", "dc1" ],
              [ "chooseleaf", "firstn", 0, "type", "host" ],
              [ "emit" ]
            ]
          }
        }

        """
        c = LibCrush(verbose=1)
        assert c.parse(json.loads(map))
        assert c.map(rule="data",
                     value=1234,
                     replication_count=1) == ["device1"]
        assert c.map(rule="data",
                     value=1234,
                     replication_count=2) == ["device1", "device5"]

    def test_map_bad(self):
        map = """
        {
          "trees": {
            "dc1": {
              "~type~": "root",
              "~id~": -1,
              "host0": {
                "~type~": "host",
                "~id~": -2,
                "device0": { "~id~": 0, "~weight~": 1.0 },
                "device1": { "~id~": 1, "~weight~": 2.0 }
              }
            }
          },
          "rules": {
            "firstn": [
              [ "take", "dc1" ],
              [ "chooseleaf", "firstn", 0, "type", "host" ],
              [ "emit" ]
            ],
            "indep": [
              [ "take", "dc1" ],
              [ "chooseleaf", "indep", 0, "type", "host" ],
              [ "emit" ]
            ]
          }
        }

        """
        c = LibCrush(verbose=1)
        assert c.parse(json.loads(map))
        assert c.map(rule="firstn",
                     value=1234,
                     replication_count=2) == ["device1"]
        assert c.map(rule="indep",
                     value=1234,
                     replication_count=2) == ["device1", None]

    def test_map_missing_map(self):
        with pytest.raises(RuntimeError) as e:
            LibCrush().map(rule="data", value=1234, replication_count=2)
        assert 'call parse()' in str(e.value)

    def test_map_no_rule(self):
        c = LibCrush(verbose=1)
        assert c.parse({})
        with pytest.raises(RuntimeError) as e:
            c.map(rule="norule",
                  value=1234,
                  replication_count=1)
        assert 'norule is not found' in str(e.value)

    def test_map_bad_replication_count(self):
        c = LibCrush(verbose=1)
        assert c.parse({
            "rules": {"data": []}
        })
        with pytest.raises(RuntimeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=0)
        assert 'must be >= 1' in str(e.value)

    def test_map_wrong_weights(self):
        c = LibCrush(verbose=1)
        assert c.parse({
            "rules": {"data": []}
        })
        with pytest.raises(RuntimeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=1,
                  weights={"unknowndevice": 1.0})
        assert 'unknowndevice is not a known device' in str(e.value)

        assert c.parse({
            "trees": {"dc1": {
                "~type~": "root",
                "device0": {
                    "~id~": 0
                }
            }},
            "rules": {"data": []}
        })
        with pytest.raises(TypeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=1,
                  weights={"device0": "abc"})

    def test_map_fail(self):
        map = """
        {
          "trees": {
            "dc1": {
              "~type~": "root",
              "~id~": -1,
              "host0": {
                "~type~": "host",
                "~id~": -2
              }
            }
          },
          "rules": {
            "data": [
              [ "take", "dc1" ],
              [ "chooseleaf", "firstn", 0, "type", "host" ],
              [ "emit" ]
            ]
          }
        }

        """
        c = LibCrush(verbose=1)
        assert c.parse(json.loads(map))
        with pytest.raises(RuntimeError) as e:
            c.map(rule="data",
                  value=1234,
                  replication_count=1)
        assert 'unable to map' in str(e.value)

# Local Variables:
# compile-command: "cd .. ; tox -e py27 tests/test_libcrush.py"
# End:



#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see `<http://www.gnu.org/licenses/>`.
#
import setuptools

setuptools.setup(
    setup_requires=['d2to1', 'pbr'],
    d2to1=True)



import json
from crush import Crush

crushmap = """
{
  "trees": {
    "dc1": {
      "~type~": "root", "~id~": -1,
      "host0": {
        "~type~": "host", "~id~": -2,
        "device0": { "~id~": 0, "~weight~": 1.0 },
        "device1": { "~id~": 1, "~weight~": 2.0 }
      },
      "host1": {
        "~type~": "host", "~id~": -3,
        "device2": { "~id~": 2, "~weight~": 1.0 },
        "device3": { "~id~": 3, "~weight~": 2.0 }
      },
      "host2": {
        "~type~": "host", "~id~": -4,
        "device4": { "~id~": 4, "~weight~": 1.0 },
        "device5": { "~id~": 5, "~weight~": 2.0 }
      }
    }
  },
  "rules": {
    "data": [
      [ "take", "dc1" ],
      [ "chooseleaf", "firstn", 0, "type", "host" ],
      [ "emit" ]
    ]
  }
}

"""

c = Crush()
c.parse(json.loads(crushmap))
print c.map(rule="data", value=1234, replication_count=1)
print c.map(rule="data", value=1234, replication_count=2)
